using System;
using System .Collections .Generic;
using System .Drawing;
using System .Drawing .Drawing2D;
using System .Windows .Forms;

using MoveGraphLibrary;

namespace Test_MoveGraphLibrary
{
    public class DoughnutSet : GraphicalObject
    {
        Point ptCenter;
        List<DoughnutRing> rings = new List<DoughnutRing> ();

        Title title;
        bool bShowTitle;
        int nTitleMiddleAngle;
        int nAddTitleRadius;

        int nMinInRadius = 20;   // min size is set to avoid disappearence while resizing
        int nMinWidth = 10;   // min size is set to avoid disappearence while resizing
        int nMinSpace = 10;
        Rectangle rcFrame;      // calculated while executing Draw()

        // -------------------------------------------------        DoughnutSet
        public DoughnutSet (Point ptC, int nInside, int nOutside, double [] values, Font fnt)
        {
            // FIRST: upgrade input values to what they have to be
            nInside = Math .Max (Math .Abs (nInside), nMinInRadius);
            nOutside = Math .Max (Math .Abs (nOutside), nInside + nMinWidth);
            if (values == null || values .Length < 2)
            {
                values = new double [] { 1, 2, 3, 4, 5 };
            }
            int nSector = values .Length;

            // NOW define the new DoughnutRing
            DoughnutRing ring = new DoughnutRing (ptC, nInside, nOutside, values, fnt);
            rings = new List<DoughnutRing> ();
            rings .Add (ring);

            ptCenter = ptC;
            bShowTitle = true;
            nTitleMiddleAngle = -90;          // place title below the circle
            nAddTitleRadius = 20;
            int nR = rings [0] .OuterRadius;
            Point pt = Auxi_Geometry .EllipseBorder (ptCenter, nR, nR,
                                                     nTitleMiddleAngle, 1.0 + (double) nAddTitleRadius / nR);
            title = new Title (pt);
        }
        // -------------------------------------------------        CalcTitlePosition
        private void CalcTitlePosition ()
        {
            int nR = OuterRadius;
            nAddTitleRadius = Math .Max (nAddTitleRadius, -nR);
            title .Anchor = Auxi_Geometry .EllipseBorder (ptCenter, nR, nR,
                                                          nTitleMiddleAngle, 1.0 + (double) nAddTitleRadius / nR);
        }
        // -------------------------------------------------        AddRing
        public void AddRing (int nInside, int nOutside, double [] values, Font fnt)
        {
            nInside = Math .Max (OuterRadius + nMinSpace, nInside);
            nOutside = Math .Max (nInside + nMinWidth, nOutside);
            DoughnutRing ring = new DoughnutRing (ptCenter, nInside, nOutside, values, fnt);
            rings .Add (ring);
            CalcTitlePosition ();

            DefineContour ();
        }
        // -------------------------------------------------        DeleteRing
        public bool DeleteRing (int iDel)
        {
            if (0 <= iDel && iDel < rings .Count && rings .Count > 1)
            {
                rings .RemoveAt (iDel);
                CalcTitlePosition ();
                DefineContour ();
                return (true);
            }
            else
            {
                return (false);
            }
        }
        // -------------------------------------------------        Draw
        public void Draw (Graphics grfx)
        {
            foreach (DoughnutRing ring in rings)
            {
                ring .Draw (grfx);
            }
            rcFrame = rings [rings .Count - 1] .Frame;
            if (bShowTitle)
            {
                title .Draw (grfx);
                rcFrame = Rectangle .Union (rcFrame, title .Frame);
            }
        }
        // -------------------------------------------------        RingCount
        public int RingCount
        {
            get { return (rings .Count); }
        }
        // -------------------------------------------------        Ring
        public DoughnutRing Ring (int i)
        {
            return (rings [Math .Min (Math .Max (0, i), RingCount - 1)]);
        }
        // -------------------------------------------------        OuterRadius
        public int OuterRadius
        {
            get { return (rings [rings .Count - 1] .OuterRadius); }
            set
            {
                if (rings [rings .Count - 1] .InnerRadius + nMinWidth <= value)
                {
                    rings [rings .Count - 1] .OuterRadius = value;
                }
            }
        }
        // -------------------------------------------------        InsideMainArea
        public bool InsideMainArea (Point pt)
        {
            foreach (DoughnutRing ring in rings)
            {
                if (ring .InsideMainArea (pt)) return (true);
            }
            return (false);
        }
        // -------------------------------------------------        ShowTitle
        public bool ShowTitle
        {
            get { return (bShowTitle); }
            set { bShowTitle = value; }
        }
        // -------------------------------------------------        MinimumSpace
        public int MinimumSpace
        {
            get { return (nMinSpace); }
            set { nMinSpace = Math .Max (0, value); }
        }
        // -------------------------------------------------        MinimumWidth
        public int MinimumWidth
        {
            get { return (nMinWidth); }
            set { nMinWidth = Math .Max (10, value); }
        }
        // -------------------------------------------------        SectorsPerRing
        public List<int> SectorsPerRing
        {
            get
            {
                List<int> nums = new List<int> ();
                foreach (DoughnutRing ring in rings)
                {
                    nums .Add (ring .Sectors);
                }
                return (nums);
            }
        }
        // -------------------------------------------------        Radiuses
        public List<int> Radiuses
        {
            get
            {
                List<int> nums = new List<int> ();
                foreach (DoughnutRing ring in rings)
                {
                    nums .Add (ring .InnerRadius);
                    nums .Add (ring .OuterRadius);
                }
                return (nums);
            }
        }
        // -------------------------------------------------        InnerRadiuses
        public List<int> InnerRadiuses
        {
            get
            {
                List<int> nums = new List<int> ();
                foreach (DoughnutRing ring in rings)
                {
                    nums .Add (ring .InnerRadius);
                }
                return (nums);
            }
        }
        // -------------------------------------------------        OuterRadiuses
        public List<int> OuterRadiuses
        {
            get
            {
                List<int> nums = new List<int> ();
                foreach (DoughnutRing ring in rings)
                {
                    nums .Add (ring .OuterRadius);
                }
                return (nums);
            }
        }

        // -------------------------------------------------        DefineContour
        public override void DefineContour ()
        {
            int nSectorsTotal = 0;
            for (int i = 0; i < rings .Count; i++)
            {
                nSectorsTotal += rings [i] .Sectors;
            }
            ContourApex [] ca = new ContourApex [nSectorsTotal * 2];
            int k = 0;
            for (int i = 0; i < rings .Count; i++)
            {
                int nRad_In = rings [i] .InnerRadius;
                int nRad_Out = rings [i] .OuterRadius;
                Rectangle rcIn = new Rectangle (ptCenter .X - nRad_In, ptCenter .Y - nRad_In, 2 * nRad_In, 2 * nRad_In);
                Rectangle rcOut = new Rectangle (ptCenter .X - nRad_Out, ptCenter .Y - nRad_Out, 2 * nRad_Out, 2 * nRad_Out);
                double [] fStartingAngles = rings [i] .StartingAngles;
                int nSectors = rings [i] .Sectors;
                for (int j = 0; j < nSectors; j++)
                {
                    float fStart = (float) (fStartingAngles [j]);
                    Point pt0 = Auxi_Geometry .EllipseBorder (rcIn, fStart, 1);
                    Point pt1 = Auxi_Geometry .EllipseBorder (rcOut, fStart, 1);

                    ca [k] = new ContourApex (k, pt0, new Size (0, 0), MovementFreedom .Any, Cursors .Hand);
                    k++;
                    ca [k] = new ContourApex (k, pt1, new Size (0, 0), MovementFreedom .Any, Cursors .Hand);
                    k++;
                }
            }
            ContourConnection [] cc = new ContourConnection [nSectorsTotal];
            for (int i = 0; i < nSectorsTotal; i++)
            {
                cc [i] = new ContourConnection (i * 2, i * 2 + 1);
            }
            contour = new Contour (ca, cc);
        }
        // -------------------------------------------------
        public override void Move (int cx, int cy)
        {
            Size size = new Size (cx, cy);
            ptCenter += size;
            foreach (DoughnutRing ring in rings)
            {
                ring .Center = ptCenter;
            }
            title .Move (cx, cy);
        }
        // -------------------------------------------------        MoveContourPoint
        public override bool MoveContourPoint (int i, int cx, int cy, Point ptM, MouseButtons catcher)
        {
            bool bRet = false;

            List<int> nSectorPerRing = SectorsPerRing;
            List<int> nInnerRadiuses = InnerRadiuses;
            List<int> nOuterRadiuses = OuterRadiuses;
            int jConnection = i / 2;
            int jRing;
            int iSectorInRing = 0;
            int iStartRing = 0;
            int nSum = 0;
            for (jRing = 0; jRing < RingCount; jRing++)
            {
                nSum += nSectorPerRing [jRing];
                if (jConnection < nSum)
                {
                    iSectorInRing = (i - iStartRing) / 2;
                    break;
                }
                iStartRing += 2 * nSectorPerRing [jRing];
            }
            double [] fStart_Deg = rings [jRing] .StartingAngles;

            if (catcher == MouseButtons .Left)
            {
                int nRad_In = nInnerRadiuses [jRing];
                int nRad_Out = nOuterRadiuses [jRing];
                Rectangle rcIn = new Rectangle (ptCenter .X - nRad_In, ptCenter .Y - nRad_In, 2 * nRad_In, 2 * nRad_In);
                Rectangle rcOut = new Rectangle (ptCenter .X - nRad_Out, ptCenter .Y - nRad_Out, 2 * nRad_Out, 2 * nRad_Out);
                // changing the size; it can be for Inner or for Outer ellipse
                Point pt, ptNew;
                int nRadNew;
                if (i % 2 == 0)     // Inner radius
                {
                    pt = Auxi_Geometry .EllipseBorder (rcIn, fStart_Deg [iSectorInRing], 1.0);
                    ptNew = new Point (pt .X + cx, pt .Y + cy);
                    nRadNew = Convert .ToInt32 (Auxi_Geometry .Distance (ptCenter, ptNew));
                    if (jRing == 0)
                    {
                        if (nMinInRadius <= nRadNew && nRadNew <= nOuterRadiuses [jRing] - nMinWidth)
                        {
                            rings [jRing] .InnerRadius = nRadNew;
                        }
                    }
                    else
                    {
                        if (nOuterRadiuses [jRing - 1] + nMinSpace <= nRadNew &&
                            nRadNew <= nOuterRadiuses [jRing] - nMinWidth)
                        {
                            rings [jRing] .InnerRadius = nRadNew;
                        }
                    }
                }
                else
                {
                    pt = Auxi_Geometry .EllipseBorder (rcOut, fStart_Deg [iSectorInRing], 1.0);
                    ptNew = new Point (pt .X + cx, pt .Y + cy);
                    nRadNew = Convert .ToInt32 (Auxi_Geometry .Distance (ptCenter, ptNew));
                    if (jRing < RingCount - 1)
                    {
                        if (nInnerRadiuses [jRing] + nMinWidth <= nRadNew &&
                            nRadNew <= nInnerRadiuses [jRing + 1] - nMinSpace)
                        {
                            rings [jRing] .OuterRadius = nRadNew;
                        }
                    }
                    else
                    {
                        if (nInnerRadiuses [jRing] + nMinWidth <= nRadNew)
                        {
                            rings [jRing] .OuterRadius = nRadNew;
                            CalcTitlePosition ();
                        }
                    }
                }
            }
            else if (catcher == MouseButtons .Right)
            {
                // Rotation
                double angleNew_Radian = -Math .Atan2 (ptM .Y - ptCenter .Y, ptM .X - ptCenter .X);
                double angleNew_Deg = angleNew_Radian * 180 / Math .PI;
                if (Math .Abs (angleNew_Deg - fStart_Deg [iSectorInRing]) > 1.0)
                {
                    int nAddAngle = Convert .ToInt32 (angleNew_Deg - fStart_Deg [iSectorInRing]);
                    rings [jRing] .StartingAngle = Auxi_Common .ChangedAngle (rings [jRing] .StartingAngle, nAddAngle);
                    DefineContour ();
                }
            }
            return (bRet);
        }

    }
}
